"""Implementation of SboTerm class.

Peter Schubert, HHU Duesseldorf, July 2022
"""
import re
import os
from functools import lru_cache
import xml.etree.ElementTree


ns = {'owl': 'http://www.w3.org/2002/07/owl#',
      'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
      'rdfs': 'http://www.w3.org/2000/01/rdf-schema#'}

local_sbo_tree = {
    'SBO_0000000': {'parents': [], 'label': 'systems biology representation'},
    'SBO_0000167': {'parents': ['SBO_0000375'], 'label': 'biochemical or transport reaction'},
    'SBO_0000176': {'parents': ['SBO_0000167'], 'label': 'biochemical reaction'},
    'SBO_0000179': {'parents': ['SBO_0000176'], 'label': 'degradation'},
    'SBO_0000180': {'parents': ['SBO_0000176'], 'label': 'dissociation'},
    'SBO_0000181': {'parents': ['SBO_0000176'], 'label': 'conformational transition'},
    'SBO_0000182': {'parents': ['SBO_0000176'], 'label': 'conversion'},
    'SBO_0000183': {'parents': ['SBO_0000205'], 'label': 'transcription'},
    'SBO_0000184': {'parents': ['SBO_0000205'], 'label': 'translation'},
    'SBO_0000204': {'parents': ['SBO_0000205'], 'label': 'DNA replication'},
    'SBO_0000205': {'parents': ['SBO_0000375'], 'label': 'composite biochemical process'},
    'SBO_0000231': {'parents': ['SBO_0000000'], 'label': 'occurring entity representation'},
    'SBO_0000236': {'parents': ['SBO_0000000'], 'label': 'physical entity representation'},
    'SBO_0000240': {'parents': ['SBO_0000236'], 'label': 'material entity'},
    'SBO_0000245': {'parents': ['SBO_0000240'], 'label': 'macromolecule'},
    'SBO_0000246': {'parents': ['SBO_0000245'], 'label': 'information macromolecule'},
    'SBO_0000247': {'parents': ['SBO_0000240'], 'label': 'simple chemical'},
    'SBO_0000250': {'parents': ['SBO_0000246'], 'label': 'ribonucleic acid'},
    'SBO_0000251': {'parents': ['SBO_0000246'], 'label': 'deoxyribonucleic acid'},
    'SBO_0000252': {'parents': ['SBO_0000246'], 'label': 'polypeptide chain'},
    'SBO_0000327': {'parents': ['SBO_0000247'], 'label': 'non-macromolecular ion'},
    'SBO_0000328': {'parents': ['SBO_0000247'], 'label': 'non-macromolecular radical'},
    'SBO_0000375': {'parents': ['SBO_0000231'], 'label': 'process'},
    'SBO_0000377': {'parents': ['SBO_0000176'], 'label': 'isomerisation'},
    'SBO_0000395': {'parents': ['SBO_0000375'], 'label': 'encapsulating process'},
    'SBO_0000589': {'parents': ['SBO_0000205'], 'label': 'genetic production'},
    'SBO_0000627': {'parents': ['SBO_0000631'], 'label': 'exchange reaction'},
    'SBO_0000628': {'parents': ['SBO_0000631'], 'label': 'demand reaction'},
    'SBO_0000629': {'parents': ['SBO_0000395'], 'label': 'biomass production'},
    'SBO_0000630': {'parents': ['SBO_0000395'], 'label': 'ATP maintenance'},
    'SBO_0000631': {'parents': ['SBO_0000375'], 'label': 'pseudoreaction'},
    'SBO_0000632': {'parents': ['SBO_0000631'], 'label': 'sink reaction'},
    'SBO_0000656': {'parents': ['SBO_0000182'], 'label': 'activation'},
    'SBO_0000672': {'parents': ['SBO_0000176'], 'label': 'spontaneous reaction'},
}
data_dir = 'data'
sbo_owl_fname = 'sbo.owl'


class SboTerm:
    """SboTerm variables

    Implemented as a flyweight pattern and caching for is_in_batch lookups
    Class builds a SBO hierachie tree, loaded from a 'sbo.owl' file,
    if file not found, tree is constructed from a local dict.
    Sbo term tree implemented as dictionary

    Class provides methods to parse the tree
    - retrieving parents
    - retrieving children
    - check if a sbo term belongs is within a specified tree branch
    - retrieving subtrees (could be used to build default tree)
    """
    _instances = {}
    sbo_tree = None

    def __init__(self, sbo_term):
        """Initialize SboTerm with sbo_term id

        :param sbo_term: SBO term id
        :type sbo_term: string in format 'SBO:0000xxx' or int
        """
        self.sbo_term = sbo_term

    def __new__(cls, sbo_term):
        """Create only one SboTerm per sbo_term.

        Flyweight pattern

        Improves performance on caching.
        Also construct initial SBO Tree

        :param sbo_term: SBO term id
        :type sbo_term: string in format 'SBO:0000xxx' or int
        """
        self = cls._instances.get(sbo_term)
        if self is None:
            self = cls._instances[sbo_term] = object.__new__(SboTerm)
            if cls.sbo_tree is None:
                cls.load_sbo_tree()
        return self

    @property
    def sbo_term(self):
        return self.__sbo_term

    @sbo_term.setter
    def sbo_term(self, sbo_term):
        if type(sbo_term) is str:
            self.__sbo_term = re.sub(':', '_', sbo_term)
        else:
            self.__sbo_term = f'SBO_{sbo_term:07d}'

    def __repr__(self):
        return self.sbo_term

    def get_parents(self):
        """Get parent hierarchy for a specific node.

        Supporting multiple parents of a node.

        :return: ordered list of unique sbo_terms of parent hierarchy
        :rtype: list of strings
        """
        return [repr(node) for node in type(self)._get_parents(self, [])]

    def get_children(self):
        """Get children subtree for a specific node.

        :return: ordered list of unique sbo_terms for children subtree
        :rtype: list of strings
        """
        return [repr(node) for node in type(self)._get_children(self, [])]

    @lru_cache
    def is_in_branch(self, upper_sbo_term):
        """Check if sbo term is part subtree starting given value.

        :param upper_sbo_term: SBO term that starts the branch
        :rtype upper_sbo_term: string or int
        :return: True if sbo_term part of the given branch
        :rtype: bool
        """
        return repr(SboTerm(upper_sbo_term)) in self.get_parents() + [self.sbo_term]

    def get_subtree(self):
        """Retrieve a subtry including parents and children of give sbo term.

        :return: all nodes forming the subtree with label and hierarchy information
        :rtype: dict with nodes forming the subtree
        """
        nodes = set(type(self)._get_children(self, [self]))
        nodes = nodes.union(set(type(self)._get_parents(self, [])))
        sub_tree = {}
        for node in nodes:
            sbo_term = node.__sbo_term
            sub_tree[sbo_term] = {'parents': type(self).sbo_tree[sbo_term]['parents'],
                                  'label': type(self).sbo_tree[sbo_term]['label']}
        return sub_tree

    @classmethod
    def _get_parents(cls, node, branch):
        """Class method to recursively identify all parents.

        :param node: sbo term of current node
        :type node: SboTerm
        :param branch: list of parents already identified
        :type branch: list of SboTerm
        :return: list of parents so far identified
        :rtype: list of SboTerm
        """
        if node.__sbo_term in SboTerm.sbo_tree:
            for parent in SboTerm.sbo_tree[node.__sbo_term]['parents']:
                sbo_parent = SboTerm(parent)
                branch.append(sbo_parent)
                cls._get_parents(sbo_parent, branch)
        return branch

    @classmethod
    def _get_children(cls, node, branch):
        """Class method to recursively identify all children.

        :param node: sbo term of current node
        :type node: SboTerm
        :param branch: list of children already identified
        :type branch: list of SboTerm
        :return: list of children so far identified
        :rtype: list of SboTerm
        """
        if node.__sbo_term in SboTerm.sbo_tree:
            for child in SboTerm.sbo_tree[node.__sbo_term]['children']:
                sbo_child = SboTerm(child)
                branch.append(sbo_child)
                cls._get_children(sbo_child, branch)
        return branch

    @classmethod
    def load_sbo_tree(cls):
        """Class method to construct an SBO Tree from file or dict.

        updated sbo file can be downloaded from https://www.ebi.ac.uk/sbo/main/download
        and stored in the package under data/sbo.owl

        """
        here = os.path.dirname(os.path.abspath(__file__))
        sbo_owl_path = os.path.join(here, data_dir, sbo_owl_fname)

        if os.path.exists(sbo_owl_path):
            tree = xml.etree.ElementTree.parse(sbo_owl_path)
            root = tree.getroot()
            sbo_tree = {}
            for sbo_class in root.findall('owl:Class', ns):
                sbo_term_id = sbo_class.attrib['{' + ns['rdf'] + '}' + 'about']
                subclasses_of = []
                label = ''
                comment = ''
                for child in sbo_class:
                    if child.tag == '{' + ns['rdfs'] + '}' + 'subClassOf':
                        if ('{' + ns['rdf'] + '}' + 'resource') in child.attrib:
                            subclasses_of.append(child.attrib['{' + ns['rdf'] + '}' + 'resource'])
                        else:
                            for class_node in child:
                                if class_node.tag == '{' + ns['owl'] + '}' + 'Class':
                                    subclasses_of.append(class_node.attrib['{' + ns['rdf'] + '}' + 'about'])
                    if child.tag == '{' + ns['rdfs'] + '}' + 'label':
                        label = child.text
                    if child.tag == '{' + ns['rdfs'] + '}' + 'comment':
                        comment = child.text
                sbo_tree[sbo_term_id] = {'parents': subclasses_of, 'label': label, 'comment': comment}
        else:
            print('loading local sbo subtree')
            sbo_tree = local_sbo_tree

        # add children information, so we can parse up/downwards through the tree
        for sbo_term in sbo_tree:
            sbo_tree[sbo_term]['children'] = []
        for sbo_term in sbo_tree:
            for parent in sbo_tree[sbo_term]['parents']:
                if sbo_term not in sbo_tree[parent]['children']:
                    sbo_tree[parent]['children'].append(sbo_term)
        SboTerm.sbo_tree = sbo_tree
